@@ -1,19 +1,16 @@
-from __future__ import absolute_import
-from __future__ import unicode_literals
-
+import enum
 import itertools
 import json
 import logging
 import os
 import re
+import subprocess
 import sys
 import tempfile
 from collections import namedtuple
 from collections import OrderedDict
 from operator import attrgetter
 
-import enum
-import six
 from docker.errors import APIError
 from docker.errors import ImageNotFound
 from docker.errors import NotFound
@@ -62,10 +59,6 @@ from .utils import truncate_id
 from .utils import unique_everseen
 from compose.cli.utils import binarystr_to_unicode
 
-if six.PY2:
-    import subprocess32 as subprocess
-else:
-    import subprocess
 
 log = logging.getLogger(__name__)
 
@@ -84,6 +77,7 @@ HOST_CONFIG_KEYS = [
     'cpuset',
     'device_cgroup_rules',
     'devices',
+    'device_requests',
     'dns',
     'dns_search',
     'dns_opt',
@@ -170,7 +164,7 @@ class BuildAction(enum.Enum):
     skip = 2
 
 
-class Service(object):
+class Service:
     def __init__(
             self,
             name,
@@ -183,6 +177,7 @@ class Service(object):
             networks=None,
             secrets=None,
             scale=1,
+            ipc_mode=None,
             pid_mode=None,
             default_platform=None,
             extra_labels=None,
@@ -194,6 +189,7 @@ class Service(object):
         self.use_networking = use_networking
         self.links = links or []
         self.volumes_from = volumes_from or []
+        self.ipc_mode = ipc_mode or IpcMode(None)
         self.network_mode = network_mode or NetworkMode(None)
         self.pid_mode = pid_mode or PidMode(None)
         self.networks = networks or {}
@@ -235,10 +231,10 @@ class Service(object):
         """Return a :class:`compose.container.Container` for this service. The
         container must be active, and match `number`.
         """
-        for container in self.containers(labels=['{0}={1}'.format(LABEL_CONTAINER_NUMBER, number)]):
+        for container in self.containers(labels=['{}={}'.format(LABEL_CONTAINER_NUMBER, number)]):
             return container
 
-        raise ValueError("No container found for %s_%s" % (self.name, number))
+        raise ValueError("No container found for {}_{}".format(self.name, number))
 
     def start(self, **options):
         containers = self.containers(stopped=True)
@@ -395,9 +391,12 @@ class Service(object):
             platform = self.default_platform
         return platform
 
-    def convergence_plan(self, strategy=ConvergenceStrategy.changed):
+    def convergence_plan(self, strategy=ConvergenceStrategy.changed, one_off=False):
         containers = self.containers(stopped=True)
 
+        if one_off:
+            return ConvergencePlan('one_off', [])
+
         if not containers:
             return ConvergencePlan('create', [])
 
@@ -413,7 +412,7 @@ class Service(object):
         stopped = [c for c in containers if not c.is_running]
 
         if stopped:
-            return ConvergencePlan('start', stopped)
+            return ConvergencePlan('start', containers)
 
         return ConvergencePlan('noop', containers)
 
@@ -425,7 +424,7 @@ class Service(object):
         except NoSuchImageError as e:
             log.debug(
                 'Service %s has diverged: %s',
-                self.name, six.text_type(e),
+                self.name, str(e),
             )
             return True
 
@@ -446,25 +445,37 @@ class Service(object):
 
         return has_diverged
 
-    def _execute_convergence_create(self, scale, detached, start):
+    def _execute_convergence_create(self, scale, detached, start, one_off=False, override_options=None):
 
         i = self._next_container_number()
 
         def create_and_start(service, n):
-            container = service.create_container(number=n, quiet=True)
+            if one_off:
+                container = service.create_container(one_off=True, quiet=True, **override_options)
+            else:
+                container = service.create_container(number=n, quiet=True)
             if not detached:
                 container.attach_log_stream()
-            if start:
+            if start and not one_off:
                 self.start_container(container)
             return container
 
+        def get_name(service_name):
+            if one_off:
+                return "_".join([
+                    service_name.project,
+                    service_name.service,
+                    "run",
+                ])
+            return self.get_container_name(service_name.service, service_name.number)
+
         containers, errors = parallel_execute(
             [
                 ServiceName(self.project, self.name, index)
                 for index in range(i, i + scale)
             ],
             lambda service_name: create_and_start(self, service_name.number),
-            lambda service_name: self.get_container_name(service_name.service, service_name.number),
+            get_name,
             "Creating"
         )
         for error in errors.values():
@@ -504,8 +515,9 @@ class Service(object):
             self._downscale(containers[scale:], timeout)
             containers = containers[:scale]
         if start:
+            stopped = [c for c in containers if not c.is_running]
             _, errors = parallel_execute(
-                containers,
+                stopped,
                 lambda c: self.start_container_if_stopped(c, attach_logs=not detached, quiet=True),
                 lambda c: c.name,
                 "Starting",
@@ -535,16 +547,20 @@ class Service(object):
     def execute_convergence_plan(self, plan, timeout=None, detached=False,
                                  start=True, scale_override=None,
                                  rescale=True, reset_container_image=False,
-                                 renew_anonymous_volumes=False):
+                                 renew_anonymous_volumes=False, override_options=None):
         (action, containers) = plan
         scale = scale_override if scale_override is not None else self.scale_num
         containers = sorted(containers, key=attrgetter('number'))
 
         self.show_scale_warnings(scale)
 
-        if action == 'create':
+        if action in ['create', 'one_off']:
             return self._execute_convergence_create(
-                scale, detached, start
+                scale,
+                detached,
+                start,
+                one_off=(action == 'one_off'),
+                override_options=override_options
             )
 
         # The create action needs always needs an initial scale, but otherwise,
@@ -628,7 +644,7 @@ class Service(object):
             expl = binarystr_to_unicode(ex.explanation)
             if "driver failed programming external connectivity" in expl:
                 log.warn("Host is already in use by another container")
-            raise OperationFailedError("Cannot start service %s: %s" % (self.name, expl))
+            raise OperationFailedError("Cannot start service {}: {}".format(self.name, expl))
         return container
 
     @property
@@ -701,32 +717,36 @@ class Service(object):
             'volumes_from': [
                 (v.source.name, v.mode)
                 for v in self.volumes_from if isinstance(v.source, Service)
-            ],
+            ]
         }
 
     def get_dependency_names(self):
         net_name = self.network_mode.service_name
         pid_namespace = self.pid_mode.service_name
+        ipc_namespace = self.ipc_mode.service_name
         return (
                 self.get_linked_service_names() +
                 self.get_volumes_from_names() +
                 ([net_name] if net_name else []) +
                 ([pid_namespace] if pid_namespace else []) +
+                ([ipc_namespace] if ipc_namespace else []) +
                 list(self.options.get('depends_on', {}).keys())
         )
 
     def get_dependency_configs(self):
         net_name = self.network_mode.service_name
         pid_namespace = self.pid_mode.service_name
+        ipc_namespace = self.ipc_mode.service_name
 
-        configs = dict(
-            [(name, None) for name in self.get_linked_service_names()]
+        configs = {
+            name: None for name in self.get_linked_service_names()
+        }
+        configs.update(
+            (name, None) for name in self.get_volumes_from_names()
         )
-        configs.update(dict(
-            [(name, None) for name in self.get_volumes_from_names()]
-        ))
         configs.update({net_name: None} if net_name else {})
         configs.update({pid_namespace: None} if pid_namespace else {})
+        configs.update({ipc_namespace: None} if ipc_namespace else {})
         configs.update(self.options.get('depends_on', {}))
         for svc, config in self.options.get('depends_on', {}).items():
             if config['condition'] == CONDITION_STARTED:
@@ -845,9 +865,9 @@ class Service(object):
         add_config_hash = (not one_off and not override_options)
         slug = generate_random_id() if one_off else None
 
-        container_options = dict(
-            (k, self.options[k])
-            for k in DOCKER_CONFIG_KEYS if k in self.options)
+        container_options = {
+            k: self.options[k]
+            for k in DOCKER_CONFIG_KEYS if k in self.options}
         override_volumes = override_options.pop('volumes', [])
         container_options.update(override_options)
 
@@ -939,7 +959,7 @@ class Service(object):
         )
         container_options['environment'].update(affinity)
 
-        container_options['volumes'] = dict((v.internal, {}) for v in container_volumes or {})
+        container_options['volumes'] = {v.internal: {} for v in container_volumes or {}}
         if version_gte(self.client.api_version, '1.30'):
             override_options['mounts'] = [build_mount(v) for v in container_mounts] or None
         else:
@@ -975,7 +995,7 @@ class Service(object):
         blkio_config = convert_blkio_config(options.get('blkio_config', None))
         log_config = get_log_config(logging_dict)
         init_path = None
-        if isinstance(options.get('init'), six.string_types):
+        if isinstance(options.get('init'), str):
             init_path = options.get('init')
             options['init'] = True
 
@@ -997,6 +1017,7 @@ class Service(object):
             privileged=options.get('privileged', False),
             network_mode=self.network_mode.mode,
             devices=options.get('devices'),
+            device_requests=options.get('device_requests'),
             dns=options.get('dns'),
             dns_opt=options.get('dns_opt'),
             dns_search=options.get('dns_search'),
@@ -1013,7 +1034,7 @@ class Service(object):
             read_only=options.get('read_only'),
             pid_mode=self.pid_mode.mode,
             security_opt=security_opt,
-            ipc_mode=options.get('ipc'),
+            ipc_mode=self.ipc_mode.mode,
             cgroup_parent=options.get('cgroup_parent'),
             cpu_quota=options.get('cpu_quota'),
             shm_size=options.get('shm_size'),
@@ -1109,7 +1130,7 @@ class Service(object):
         try:
             all_events = list(stream_output(build_output, output_stream))
         except StreamOutputError as e:
-            raise BuildError(self, six.text_type(e))
+            raise BuildError(self, str(e))
 
         # Ensure the HTTP connection is not reused for another
         # streaming command, as the Docker daemon can sometimes
@@ -1141,9 +1162,9 @@ class Service(object):
     def labels(self, one_off=False, legacy=False):
         proj_name = self.project if not legacy else re.sub(r'[_-]', '', self.project)
         return [
-            '{0}={1}'.format(LABEL_PROJECT, proj_name),
-            '{0}={1}'.format(LABEL_SERVICE, self.name),
-            '{0}={1}'.format(LABEL_ONE_OFF, "True" if one_off else "False"),
+            '{}={}'.format(LABEL_PROJECT, proj_name),
+            '{}={}'.format(LABEL_SERVICE, self.name),
+            '{}={}'.format(LABEL_ONE_OFF, "True" if one_off else "False"),
         ]
 
     @property
@@ -1160,7 +1181,7 @@ class Service(object):
         ext_links_origins = [link.split(':')[0] for link in self.options.get('external_links', [])]
         if container_name in ext_links_origins:
             raise DependencyError(
-                'Service {0} has a self-referential external link: {1}'.format(
+                'Service {} has a self-referential external link: {}'.format(
                     self.name, container_name
                 )
             )
@@ -1215,16 +1236,14 @@ class Service(object):
             output = self.client.pull(repo, **pull_kwargs)
             if silent:
                 with open(os.devnull, 'w') as devnull:
-                    for event in stream_output(output, devnull):
-                        yield event
+                    yield from stream_output(output, devnull)
             else:
-                for event in stream_output(output, sys.stdout):
-                    yield event
+                yield from stream_output(output, sys.stdout)
         except (StreamOutputError, NotFound) as e:
             if not ignore_pull_failures:
                 raise
             else:
-                log.error(six.text_type(e))
+                log.error(str(e))
 
     def pull(self, ignore_pull_failures=False, silent=False, stream=False):
         if 'image' not in self.options:
@@ -1237,7 +1256,7 @@ class Service(object):
             'platform': self.platform,
         }
         if not silent:
-            log.info('Pulling %s (%s%s%s)...' % (self.name, repo, separator, tag))
+            log.info('Pulling {} ({}{}{})...'.format(self.name, repo, separator, tag))
 
         if kwargs['platform'] and version_lt(self.client.api_version, '1.35'):
             raise OperationFailedError(
@@ -1255,7 +1274,7 @@ class Service(object):
 
         repo, tag, separator = parse_repository_tag(self.options['image'])
         tag = tag or 'latest'
-        log.info('Pushing %s (%s%s%s)...' % (self.name, repo, separator, tag))
+        log.info('Pushing {} ({}{}{})...'.format(self.name, repo, separator, tag))
         output = self.client.push(repo, tag=tag, stream=True)
 
         try:
@@ -1265,7 +1284,7 @@ class Service(object):
             if not ignore_push_failures:
                 raise
             else:
-                log.error(six.text_type(e))
+                log.error(str(e))
 
     def is_healthy(self):
         """ Check that all containers for this service report healthy.
@@ -1310,6 +1329,24 @@ class Service(object):
 
         return result
 
+    def get_profiles(self):
+        if 'profiles' not in self.options:
+            return []
+
+        return self.options.get('profiles')
+
+    def enabled_for_profiles(self, enabled_profiles):
+        # if service has no profiles specified it is always enabled
+        if 'profiles' not in self.options:
+            return True
+
+        service_profiles = self.options.get('profiles')
+        for profile in enabled_profiles:
+            if profile in service_profiles:
+                return True
+
+        return False
+
 
 def short_id_alias_exists(container, network):
     aliases = container.get(
@@ -1317,7 +1354,47 @@ def short_id_alias_exists(container, network):
     return container.short_id in aliases
 
 
-class PidMode(object):
+class IpcMode:
+    def __init__(self, mode):
+        self._mode = mode
+
+    @property
+    def mode(self):
+        return self._mode
+
+    @property
+    def service_name(self):
+        return None
+
+
+class ServiceIpcMode(IpcMode):
+    def __init__(self, service):
+        self.service = service
+
+    @property
+    def service_name(self):
+        return self.service.name
+
+    @property
+    def mode(self):
+        containers = self.service.containers()
+        if containers:
+            return 'container:' + containers[0].id
+
+        log.warning(
+            "Service %s is trying to use reuse the IPC namespace "
+            "of another service that is not running." % (self.service_name)
+        )
+        return None
+
+
+class ContainerIpcMode(IpcMode):
+    def __init__(self, container):
+        self.container = container
+        self._mode = 'container:{}'.format(container.id)
+
+
+class PidMode:
     def __init__(self, mode):
         self._mode = mode
 
@@ -1357,7 +1434,7 @@ class ContainerPidMode(PidMode):
         self._mode = 'container:{}'.format(container.id)
 
 
-class NetworkMode(object):
+class NetworkMode:
     """A `standard` network mode (ex: host, bridge)"""
 
     service_name = None
@@ -1372,7 +1449,7 @@ class NetworkMode(object):
     mode = id
 
 
-class ContainerNetworkMode(object):
+class ContainerNetworkMode:
     """A network mode that uses a container's network stack."""
 
     service_name = None
@@ -1389,7 +1466,7 @@ class ContainerNetworkMode(object):
         return 'container:' + self.container.id
 
 
-class ServiceNetworkMode(object):
+class ServiceNetworkMode:
     """A network mode that uses a service's network stack."""
 
     def __init__(self, service):
@@ -1494,10 +1571,10 @@ def get_container_data_volumes(container, volumes_option, tmpfs_option, mounts_o
     volumes = []
     volumes_option = volumes_option or []
 
-    container_mounts = dict(
-        (mount['Destination'], mount)
+    container_mounts = {
+        mount['Destination']: mount
         for mount in container.get('Mounts') or {}
-    )
+    }
 
     image_volumes = [
         VolumeSpec.parse(volume)
@@ -1549,9 +1626,9 @@ def get_container_data_volumes(container, volumes_option, tmpfs_option, mounts_o
 
 
 def warn_on_masked_volume(volumes_option, container_volumes, service):
-    container_volumes = dict(
-        (volume.internal, volume.external)
-        for volume in container_volumes)
+    container_volumes = {
+        volume.internal: volume.external
+        for volume in container_volumes}
 
     for volume in volumes_option:
         if (
@@ -1631,8 +1708,8 @@ def build_ulimits(ulimit_config):
     if not ulimit_config:
         return None
     ulimits = []
-    for limit_name, soft_hard_values in six.iteritems(ulimit_config):
-        if isinstance(soft_hard_values, six.integer_types):
+    for limit_name, soft_hard_values in ulimit_config.items():
+        if isinstance(soft_hard_values, int):
             ulimits.append({'name': limit_name, 'soft': soft_hard_values, 'hard': soft_hard_values})
         elif isinstance(soft_hard_values, dict):
             ulimit_dict = {'name': limit_name}
@@ -1656,7 +1733,7 @@ def format_environment(environment):
     def format_env(key, value):
         if value is None:
             return key
-        if isinstance(value, six.binary_type):
+        if isinstance(value, bytes):
             value = value.decode('utf-8')
         return '{key}={value}'.format(key=key, value=value)
 
@@ -1701,24 +1778,19 @@ def convert_blkio_config(blkio_config):
             continue
         arr = []
         for item in blkio_config[field]:
-            arr.append(dict([(k.capitalize(), v) for k, v in item.items()]))
+            arr.append({k.capitalize(): v for k, v in item.items()})
         result[field] = arr
     return result
 
 
 def rewrite_build_path(path):
-    # python2 os.stat() doesn't support unicode on some UNIX, so we
-    # encode it to a bytestring to be safe
-    if not six.PY3 and not IS_WINDOWS_PLATFORM:
-        path = path.encode('utf8')
-
     if IS_WINDOWS_PLATFORM and not is_url(path) and not path.startswith(WINDOWS_LONGPATH_PREFIX):
         path = WINDOWS_LONGPATH_PREFIX + os.path.normpath(path)
 
     return path
 
 
-class _CLIBuilder(object):
+class _CLIBuilder:
     def __init__(self, progress):
         self._progress = progress
 
@@ -1794,6 +1866,7 @@ class _CLIBuilder(object):
         command_builder.add_flag("--force-rm", forcerm)
         command_builder.add_params("--label", labels)
         command_builder.add_arg("--memory", container_limits.get("memory"))
+        command_builder.add_arg("--network", network_mode)
         command_builder.add_flag("--no-cache", nocache)
         command_builder.add_arg("--progress", self._progress)
         command_builder.add_flag("--pull", pull)
@@ -1804,18 +1877,20 @@ class _CLIBuilder(object):
 
         magic_word = "Successfully built "
         appear = False
-        with subprocess.Popen(args, stdout=subprocess.PIPE, universal_newlines=True) as p:
+        with subprocess.Popen(args, stdout=subprocess.PIPE,
+                              universal_newlines=True) as p:
             while True:
                 line = p.stdout.readline()
                 if not line:
                     break
-                # Fix non ascii chars on Python2. To remove when #6890 is complete.
-                if six.PY2:
-                    magic_word = str(magic_word)
                 if line.startswith(magic_word):
                     appear = True
                 yield json.dumps({"stream": line})
 
+            p.communicate()
+            if p.returncode != 0:
+                raise StreamOutputError()
+
         with open(iidfile) as f:
             line = f.readline()
             image_id = line.split(":")[1].strip()
@@ -1829,7 +1904,7 @@ class _CLIBuilder(object):
             yield json.dumps({"stream": "{}{}\n".format(magic_word, image_id)})
 
 
-class _CommandBuilder(object):
+class _CommandBuilder:
     def __init__(self):
         self._args = ["docker", "build"]
 
